import {buildTreeStruct, getAlternative, serialize} from "./dictionary-tree.mjs";
import {parse} from "./dictionary-parser.mjs";

export class Dictionary {
	constructor(dictionaryString, options= {}) {
		this.originalConfig = options;
		this.originalConfig.dictionaryString = dictionaryString;
		this.config = JSON.parse(JSON.stringify(this.originalConfig));
		if(typeof this.config.cache === 'undefined') this.config.cache = true;
		if(typeof this.config.cacheMax === 'undefined') this.config.cacheMax = 1_000_000;
		if(typeof this.config.idSecPwd === 'undefined') this.config.idSecPwd = true;
		this.tree = buildTreeStruct(parse(dictionaryString,this.config.idSecPwd));
		this.length = this.tree.altCount;
		this.estimateDuration = ()=>this.length/(this.config.speed || 1)
		this.estimateRemaining = ()=>(this.length-this.tried)/(this.config.speed || 1)
		if(this.config.speed) adjustVariant(this);
		this.tried = 0;
		this.cache = [];
		this.duplicatedCount = ()=> Object.keys(this.cache).length? this.tried - Object.keys(this.cache).length : 0;
		this.dryGet = index => getAlternative(index,this.tree);
		this.get = index => {
			if(typeof this.startTime === "undefined") this.startTime = Date.now();
			this.tried++;
			const alt = getAlternative(index,this.tree);
			if(this.config.cache && this.length < this.config.cacheMax) {
				if(typeof this.cache[alt] === "undefined") this.cache[alt] = [];
				this.cache[alt].push(index);
				if(this.cache[alt].length>1) return `§duplicate§${alt}`;
			}
			return alt;
		}
		this.timeSpent = ()=>(Date.now()-this.startTime)/1000;
		this.duplicatedFound = ()=>{
			const duplicated = [];
			for(let key of Object.keys(this.cache)) if(this.cache[key].length>1) duplicated.push({alt:key,index:this.cache[key]});
			const sortedDuplicate = duplicated.sort((a,b)=>b.index.length-a.index.length);
			return sortedDuplicate;
		}
		this.dryRunDedup = ()=>dryRunDedup(this);
	}
}
function adjustVariant(self){
	const durationGoal = 3_600;
	if (self.estimateDuration() >= durationGoal) return;
	if (self.length >= self.config.cacheMax) return;
	function rebuildTree(self){
		self.tree = buildTreeStruct(parse(self.originalConfig.dictionaryString,self.config.idSecPwd,self.config.accent,self.config.lowerCase,self.config.leetSpeak));
		self.length = self.tree.altCount;
	}
	let lastLess1hConfig;
	while (self.estimateDuration() < durationGoal && self.length < self.config.cacheMax){
		lastLess1hConfig = JSON.parse(JSON.stringify(self.config));
		if(self.config.accent !== 2 && (typeof self.originalConfig.accent === 'undefined' || self.originalConfig.accent === "auto")) {
			if(!self.config.accent) self.config.accent = 1;
			else self.config.accent++;
			rebuildTree(self);
			continue;
		}
		if(self.config.lowerCase !== 4 && (typeof self.originalConfig.lowerCase === 'undefined' || self.originalConfig.lowerCase === "auto")) {
			if(!self.config.lowerCase) self.config.lowerCase = 1;
			else self.config.lowerCase++;
			rebuildTree(self);
			continue;
		}
		if(self.config.leetSpeak !== 3 && (typeof self.originalConfig.leetSpeak === 'undefined' || self.originalConfig.leetSpeak === "auto")) {
			if(!self.config.leetSpeak) self.config.leetSpeak = 1;
			else self.config.leetSpeak++;
			rebuildTree(self);
			continue;
		}
		console.log('strange, not an hour yet with all variants ?');
		break;
	}
	self.config = lastLess1hConfig || self.config;
	rebuildTree(self);
	dryRunDedup(self);
}
function dryRunDedup(self){
	const dry = new Dictionary(serialize(self.tree),{cache:true,idSecPwd:false});
	for(let i = 0; i < dry.length;i++) dry.get(i);
	const duplicate = dry.duplicatedFound();
	self.duplicateTotal = dry.duplicatedCount();
	self.duplicateRatio = self.duplicateTotal/self.length;
	self.uniqueRatio = 1 - self.duplicateRatio;
	return duplicate;
}
